c----------------------------------------------------------------------
c --- New subroutines required for FLARE-Source update to CALPUFF v6.42
c----------------------------------------------------------------------
c
c (1)   RDFL_CF
c       PURPOSE:  Read the FLARE section at the end of the CALPUFF
c                 control file, and perform QA on content
c
c (2)   RDTIFL2 
c       PURPOSE:  Read time-invariant header records of a FLEMARB
c                 FLARE source file and store in corresponding PTEMARB
c                 arrays
c
c (3)   SKIPHDFL2
c       PURPOSE:  Use null reads to skip header records in one FLEMARB
c                 FLARE file
c
c (4)   RDFL2
c       PURPOSE:  Read time-varying data from one FLEMARB flare file
c
c (5)   SWAPTIEM2
c       PURPOSE:  Swap location information from current time-varying
c                 sources into the PT2 time-invariant storage array
c
c (6)   MK_CASE
c       PURPOSE:  Imposes uniform case on a string
c----------------------------------------------------------------------


c----------------------------------------------------------------------
      subroutine rdfl_cf
c----------------------------------------------------------------------
c
c --- CALPUFF    Version: 6.42_x1.0     Level: 121203           RDFL_CF
c
c
c --- PURPOSE:  Read the FLARE section at the end of the CALPUFF
c               control file, and perform QA on content
c
c --- INPUTS:
c           /FILNAM/
c             nptdat, ptdat
c           /PT2/
c             npt2
c
c --- OUTPUT:
c
c           /FILNAM/
c             ptdat, nfldat, nptfldat
c           /FLAGS/
c             mtip_fl, mrise_fl
c           /FL2/
c             nfl2, nptfl2
c
c --- RDFL_CF called by:  SETUP
c --- RDFL_CF calls:      MK_CASE
c----------------------------------------------------------------------
      include 'params.puf'
      include 'filnam.puf'
      include 'flags.puf'
      include 'pt2.puf'
      include 'fl2.puf'

      character*132 blank132,aline132
      character*1 dlim,eq

      data dlim/'!'/,eq/'='/

      do i=1,132
         blank132(i:i)=' '
      enddo

c --- Info to list file
      write(io6,*)
      write(io6,*)
      write(io6,*)'--------------------------------------------------'
      write(io6,*)'INPUT GROUP: FL -- FLARE source control parameters'
      write(io6,*)'--------------------------------------------------'
      write(io6,*)

c --- Process control flag section (to first END or EOF)
10    aline132=blank132
      read(io5,'(a132)',end=20) aline132
      i1=INDEX(aline132,dlim)
      i2=INDEX(aline132,eq)
      i3=INDEX(aline132(i1+1:132),dlim)
      if(i1.EQ.0 .AND. i3.EQ.0) then
c ---    No assignment here
         goto 10
      elseif(i1.GT.0 .AND. i2.EQ.0 .AND. i3.GT.0) then
c ---    Must be an END (no equal sign)
c ---    Impose upper-case
         call MK_CASE(io6,'UPPER',aline132,ierr)
         if(ierr.GT.0) stop 'HALTED in RDFL_CF -- See list file'
c ---    Reference pointer 3 to the start of the full line
         i3=i1+i3
         if(INDEX(aline132(i1:i3),'END').GT.0) then
c ---       Next section of file
            goto 20
         else
            write(io6,*)'ERROR in RDFL_CF:  Invalid assignment'
            write(io6,*)TRIM(aline132)
            stop 'HALTED in RDFL_CF -- See list file'
         endif
      elseif(i1.GT.0 .AND. i2.GT.0 .AND. i3.GT.0) then
c ---    Reference pointer 3 to the start of the full line
         i3=i1+i3
c ---    Assignment
         l1=i2+1
         l2=i3-1
c ---    Impose upper-case
         call MK_CASE(io6,'UPPER',aline132,ierr)
         if(ierr.GT.0) stop 'HALTED in RDFL_CF -- See list file'
c ---    Look for recognized variables
         if(INDEX(aline132(i1:i2),'NFL2').GT.0) then
            read(aline132(l1:l2),'(i)') nfl2
         elseif(INDEX(aline132(i1:i2),'NFLDAT').GT.0) then
            read(aline132(l1:l2),'(i)') nfldat
         elseif(INDEX(aline132(i1:i2),'MTIP_FL').GT.0) then
            read(aline132(l1:l2),'(i)') mtip_fl
         elseif(INDEX(aline132(i1:i2),'MRISE_FL').GT.0) then
            read(aline132(l1:l2),'(i)') mrise_fl
         else
            write(io6,*)'ERROR in RDFL_CF:  Assignment not recognized'
            write(io6,*)TRIM(aline132)
            stop 'HALTED in RDFL_CF -- See list file'
         endif
      else
         write(io6,*)'ERROR in RDFL_CF:  Invalid assignment'
         write(io6,*)TRIM(aline132)
         stop 'HALTED in RDFL_CF -- See list file'
      endif

c --- Get another line
      goto 10

20    nptfl2=npt2+nfl2
      nptfldat=nptdat+nfldat

c --- Info to list file
      write(io6,*)
      write(io6,*)'   MTIP_FL                  = ',mtip_fl
      write(io6,*)'   MRISE_FL                 = ',mrise_fl
      write(io6,*)'   Number of FLARE Sources  = ',nfl2
      write(io6,*)'   Number of Emission Files = ',nfldat
      write(io6,*)

c --- If there are flare sources there must be filenames
      if(nfl2.GT.0 .AND. nfldat.LE.0) then
         write(io6,*)'ERROR in RDFL_CF:  no FLARE filenames are listed'
         write(io6,*)'Number of FLARE sources = ',nfl2
         write(io6,*)'There must be filenames with FLARE sources'
         stop 'HALTED in RDFL_CF -- See list file'
      endif

c --- Check array space for combined POINTS and FLARES
      if(nptfldat.GT.MXEMDAT) then
         write(io6,*)'ERROR in RDFL_CF:  no room for FLARE file data'
         write(io6,*)'Number of FLARE files = ',nfldat
         write(io6,*)'Number of POINT files = ',nptdat
         write(io6,*)'Arrays size for total = ',MXEMDAT
         write(io6,*)'Increase MXEMDAT in PARAMS.PUF and recompile'
         stop 'HALTED in RDFL_CF -- See list file'
      endif
      if(nptfl2.GT.MXPT2) then
         write(io6,*)'ERROR in RDFL_CF:  no room for FLARE sources'
         write(io6,*)'Number of FLARE sources = ',nfl2
         write(io6,*)'Number of POINT sources = ',npt2
         write(io6,*)'Arrays size for total   = ',MXPT2
         write(io6,*)'Increase MXPT2 in PARAMS.PUF and recompile'
         stop 'HALTED in RDFL_CF -- See list file'
      endif

c --- Check FLARE flags
      if(mtip_fl.LT.0 .OR. mtip_fl.GT.1 .OR.
     &   mrise_fl.LT.1 .OR. mrise_fl.GT.2) then
         write(io6,*)'ERROR in RDFL_CF:  Invalid FLARE flags'
         write(io6,*)'Expected MRISE_FL = 1 or 2, found: ',mrise_fl
         write(io6,*)'Expected  MTIP_FL = 0 or 1, found: ',mtip_fl
         stop 'HALTED in RDFL_CF -- See list file'
      endif

c --- Done if no files
      if(nfldat.LE.0) return

c --- Process list of file names
      names=0
30    aline132=blank132
      read(io5,'(a132)',end=999) aline132
      i1=INDEX(aline132,dlim)
      i2=INDEX(aline132,eq)
      i3=INDEX(aline132(i1+1:132),dlim)
      if(i1.EQ.0 .AND. i3.EQ.0) then
c ---    No assignment here
         goto 30
      elseif(i1.GT.0 .AND. i2.GT.0 .AND. i3.GT.0) then
c ---    Reference pointer 3 to the start of the full line
         i3=i1+i3
c ---    Assignment
         l1=i2+1
         l2=i3-1
c ---    Impose upper-case
         call MK_CASE(io6,'UPPER',aline132,ierr)
c ---    Look for recognized variables
         if(INDEX(aline132(i1:i2),'FLDAT').GT.0) then
c ---       Assign name
            names=names+1
            i=nptdat+names
            ptdat(i)=aline132(l1:l2)
c ---       Left-justify
            ptdat(i)=ADJUSTL(ptdat(i))
c ---       More expected?
            if(names.LT.nfldat) goto 30
         else
            write(io6,*)'ERROR in RDFL_CF:  Assignment not recognized'
            write(io6,*)TRIM(aline132)
            stop 'HALTED in RDFL_CF -- See list file'
         endif
      else
         write(io6,*)'ERROR in RDFL_CF:  Invalid assignment'
         write(io6,*)TRIM(aline132)
         stop 'HALTED in RDFL_CF -- See list file'
      endif

c --- Report filenames for FLARES
      write(io6,*)
      write(io6,*)'   FLARE source variable emission files:'
      write(io6,*)
      do i=nptdat+1,nptfldat
         write(io6,*)'   '//TRIM(ptdat(i))
      enddo
      write(io6,*)
      write(io6,*)

      return

999   write(io6,*)'ERROR in RDFL_CF:  Unexpected EOF while reading'//
     &            ' FLARE variable emissions filenames'
      write(io6,*)'Number of filenames expected: ',nfldat
      write(io6,*)'Number of filenames found   : ',names
      stop 'HALTED in RDFL_CF -- See list file'
      end

c----------------------------------------------------------------------
      subroutine RDTIFL2(iobase,nptfl2,nfldat,nptfldat,i,ibsrc2,iesrc2,
     &                   avrs2,ldbug,io6,iox,tifl2,cid2,ncommout)
c----------------------------------------------------------------------
c
c --- CALPUFF    Version: 6.42_x1.0     Level: 121203           RDTIFL2
c                D. Strimaitis
c
c --- PURPOSE:  Read time-invariant header records of a FLEMARB
c               FLARE source file and store in corresponding PTEMARB
c               arrays
c
c --- INPUTS:
c           IOBASE - integer    - Fortran unit number for the FIRST
c                                 FLEMARB file
c           NPTFL2 - integer    - Number of PTEMARB+FLEMARB sources
c           NFLDAT - integer    - Number of FLEMARB files
c         NPTFLDAT - integer    - Number of PTEMARB+FLEMARB files
c                I - integer    - Current PTEMARB+FLEMARB file
c IBSRC2(nptfldat) - integer    - First source in each file
c                                 (PTEMARB/FLEMARB)
c IESRC2(nptfldat) - integer    - Last source in each file
c                                 (PTEMARB/FLEMARB)
c            AVRS2 - char*16    - PTEMARB/FLEMARB data-set version
c            LDBUG - logical    - Flag for debug output
c              IO6 - integer    - Fortran unit number for list file
c              IOX - integer    - Fortran unit number for scratch file
c
c --- OUTPUT:
c  TIFL2(8,nptfl2) - real array - Time-invariant FLEMARB source data
c     CID2(nptfl2) - char*16    - FLEMARB source identifiers
c
c --- INPUT/OUTPUT:
c         NCOMMOUT - integer    - Number of comments in scratch file
c
c --- RDTIFL2 called by: SETUP
c --- RDTIFL2 calls:     none
c----------------------------------------------------------------------
c
      real tifl2(8,nptfl2)
      integer ibsrc2(nptfldat),iesrc2(nptfldat)
      character*16 cid2(nptfl2),avrs2
      logical ldbug

c --- Check on dataset version (may signal array problems)
      if(TRIM(avrs2).NE.'1.0') then
         write(io6,*)'ERROR in RDTIFL2: bad dataset'
         write(io6,*)'Expected FLEMARB dataset: 1.0'
         write(io6,*)'Found                    : '//TRIM(avrs2)
         stop 'HALTED in RDTIFL2 -- See list file'
      endif

c --- Set file unit
      np=nptfldat-nfldat
      io=iobase+(i-np)-1

c --- Source Loop
      do isrc=ibsrc2(i),iesrc2(i)
         if(isrc.GT.nptfl2) then
            write(io6,*)'ERROR in RDTIFL2: too many sources'
            write(io6,*)'Expected a total of: ',nptfl2
            stop 'HALTED in RDTIFL2 -- See list file'
         endif
         read(io,*) cid2(isrc),(tifl2(n,isrc),n=1,8)
         write(iox,'(a16,f10.0)') cid2(isrc),(tifl2(n,isrc),n=1,8)
         ncommout=ncommout+1
c ---    Buoyancy and momentum flux option is not available
         if(NINT(tifl2(1,isrc)).NE.1) then
            write(io6,*)'ERROR in RDTIFL2: unsupported option'
            write(io6,*)'Buoyancy/momentum flux option not available'
            stop 'HALTED in RDTIFL2 -- See list file'
         endif
      enddo

c --- Processing when all files have been read
      if(i.EQ.nptfldat) then

c ---    Number of sources should be total expected
         if(iesrc2(i).NE.nptfl2) then
            write(io6,*)'ERROR in RDTIFL2:  Wrong number of sources'
            write(io6,*)'Expected : ',nptfl2
            write(io6,*)'Found    : ',iesrc2(i)
            write(*,*)
            stop 'HALTED in RDTIFL2 -- See list file'
         endif
      endif

      if(ldbug)then
c ---    Report data
         write(io6,*)
         write(io6,*)
         write(io6,*)'Time-invariant FLEMARB data'
         write(io6,*)'-----------------------------'
         write(io6,*)' CID2         TIFL2(1-8)'
         do isrc=ibsrc2(i),iesrc2(i)
            write(io6,*) cid2(isrc),(tifl2(n,isrc),n=1,8)
         enddo
         write(io6,*)
      endif


      return
      end

c----------------------------------------------------------------------
      subroutine SKIPHDFL2(io,iptfl)
c----------------------------------------------------------------------
c
c --- CALPUFF    Version: 6.42_x1.0     Level: 121203         SKIPHDFL2
c                D. Strimaitis
c
c --- PURPOSE:  Use null reads to skip header records in one FLEMARB
c               FLARE file
c
c --- INPUTS:
c               IO - integer   - Unit no. for current file
c            IPTFL - integer   - Current index in list of files
c                                (PTEMARB+FLEMARB)
c
c --- OUTPUT:
c
c --- SKIPHDFL2 called by: RDEMSRC
c --- SKIPHDFL2 calls:     none
c----------------------------------------------------------------------
      include 'params.puf'
      include 'pt2.puf'

      real rdata(mxspec)
      character*12 cdata(mxspec)

c --- Rewind to start at top of header
      REWIND(io)

c --- Read records with nulls
      read(io,*)
      read(io,*) ncom
      do k=1,ncom
         read(io,*)
      enddo
      read(io,*)
      read(io,*)
      if(LLCCPT2(iptfl) .OR. LLAZAPT2(iptfl) .OR. LTTMPT2(iptfl))
     &   read(io,*)
      read(io,*)
      read(io,*)
      read(io,*)
      read(io,*)
      read(io,*)
c --- Emitted Species
      read(io,*) (cdata(k),k=1,nse2)
c --- Molecular weights
      read(io,*) (rdata(k),k=1,nse2)
c --- Time-invariant records
c --- Source loop
      do k=ibsrc2(iptfl),iesrc2(iptfl)
         read(io,*)
      enddo

      return
      end

c----------------------------------------------------------------------
      subroutine RDFL2(io,nsrc,nse,cid,ibsrc,iesrc,tzhr,ldb,io6,
     &                 ibdh,ibs,iedh,ies,data,ieof)
c----------------------------------------------------------------------
c
c --- CALPUFF    Version: 6.42_x1.0     Level: 121203             RDFL2
c                D. Strimaitis
c
c --- PURPOSE:  Read time-varying data from one FLEMARB flare file
c
c --- INPUTS:
c              IO - integer    - Fortran unit number of FLEMARB file
c            NSRC - integer    - Total number of sources in all
c                                PTEMARB+FLEMARB files
c             NSE - integer    - Number of emitted species
c       CID(nsrc) - char*16    - Source names
c           IBSRC - integer    - First source in current file
c           IESRC - integer    - Last source in current file
c            TZHR - real       - Hours to ADD to times in file to
c                                convert to BASE TIME ZONE
c             LDB - logical    - Flag for debug output
c             IO6 - integer    - Fortran unit number for list file
c
c --- OUTPUT:
c        IBDH,IBS - integer    - Starting date-time for period of data
c                                (YYYYJJJHH & SSSS) BASE TIME ZONE
c        IEDH,IES - integer    - Ending date-time for period of data
c                                (YYYYJJJHH & SSSS) BASE TIME ZONE
c  DATA(NSE+11,nsrc) - real    - Time-varying data
c             IEOF - integer   - End-of-File
c
c --- RDFL2 called by: INITPUF
c --- RDFL2 calls:     YR4, INCR, TCLOCK
c----------------------------------------------------------------------

c --- Declarations
      real data(nse+11,nsrc)
      character*16 cid0, cid(nsrc)
      logical ldb, lbad

c --- Small value to screen for non-zero horizontal exit velocity
      data small/0.0001/

c --- Initializations
      ieof=0
      nvar=nse+11

c --- Date and time for data
      read(io,*,end=999) ibyr,ibdy,ibhr,ibs,ieyr,iedy,iehr,ies

c --- Year format YYYY
      call YR4(io6,ibyr,ierrb)
      call YR4(io6,ieyr,ierre)
      if(ierrb.NE.0 .OR. ierre.NE.0) then
         write(*,*)
         stop 'HALTED in RDFL2 -- See list file.'
      endif

c --- Clock time must be between 00:0000 and 23:3599
      idyhrdum=-9
      call TCLOCK(io6,idyhrdum,ibyr,ibdy,ibhr,ibs)
      call TCLOCK(io6,idyhrdum,ieyr,iedy,iehr,ies)

c --- BASE TIME ZONE
      call INCR(io6,ibyr,ibdy,ibhr,NINT(tzhr))
      call INCR(io6,ieyr,iedy,iehr,NINT(tzhr))

c --- Make date-hour
      ibdh=ibyr*100000+ibdy*100+ibhr
      iedh=ieyr*100000+iedy*100+iehr

c --- Source loop with QA
      lbad = .FALSE.
      do i=ibsrc,iesrc
         read(io,*) cid0,(data(n,i),n=1,nvar)
c ---    Check for bad source id
         if(cid0.NE.cid(i)) then
            lbad=.TRUE.
            write(io6,*)
            write(io6,*)'RDFL2: Bad source ID in FLEMARB.DAT'
            write(io6,*)'       Expected source = ',cid(i)
            write(io6,*)'       Found source    = ',cid0
            write(io6,*)
         endif
c ---    Check for non-vertical exit velocity (not treated)
         if(ABS(data(7,i)).GT.small .OR. ABS(data(8,i)).GT.small) then
            lbad=.TRUE.
            write(io6,*)
            write(io6,*)'RDFL2: Exit vel. not vertical in FLEMARB.DAT'
            write(io6,*)'       Source ',cid0
            write(io6,*)'       Expected U,V = 0.0'
            write(io6,*)'       Found    U,V = ',data(7,i),data(8,i)
            write(io6,*)
         endif
      enddo

c --- Act on data problem
      if(lbad) then
         write(io6,*)
         write(io6,*)'RDFL2: Problem with source data in FLEMARB.DAT'
         write(io6,*)'Period begins: ',ibdh,ibs
         write(io6,*)'Period ends  : ',iedh,ies
         write(*,*)
         stop 'HALTED in RDFL2 -- See list file.'
      endif

c --- Debug output
      if(ldb)then
         write(io6,*)
         write(io6,*)
         write(io6,*)'Time-varying FLEMARB data'
         write(io6,*)'Sources      : ',ibsrc,iesrc
         write(io6,*)'Period begins: ',ibdh,ibs
         write(io6,*)'Period ends  : ',iedh,ies
         do i=ibsrc,iesrc
            write(io6,*) cid(i),'  DATA = ',(data(n,i),n=1,nvar)
         enddo
      endif

      return

c --- Unexpected end-of-file
999   ieof=1
      write(io6,*)
      write(io6,*)'RDFL2: Unexpected end in FLEMARB.DAT file'
      write(*,*)
      write(*,*)  'RDFL2: Unexpected end in FLEMARB.DAT file'

      return

      end

c----------------------------------------------------------------------
      subroutine SWAPTIEM2(nsrc,nse,ibsrc2,iesrc2,data,tiem2)
c----------------------------------------------------------------------
c
c --- CALPUFF    Version: 6.42_x1.0     Level: 121203         SWAPTIEM2
c                D. Strimaitis
c
c --- PURPOSE:  Swap location information from current time-varying
c               sources into the PT2 time-invariant storage array
c
c --- INPUTS:
c            NSRC - integer    - Total number of sources in all
c                                PTEMARB+FLEMARB files
c             NSE - integer    - Number of emitted species
c           IBSRC - integer    - First source in current file
c           IESRC - integer    - Last source in current file
c  DATA(NSE+11,nsrc) - real    - Time-varying data
c
c --- OUTPUT:
c   tiem2(8,nsrc) - real       - Time-invariant PT2 data
c
c --- SWAPTIEM2 called by: SETUP
c --- SWAPTIEM2 calls:     
c----------------------------------------------------------------------

c --- Declarations
      real data(nse+11,nsrc)
      real tiem2(8,nsrc)

      do is=ibsrc2,iesrc2
         tiem2(1,is)=data(1,is)         ! x
         tiem2(2,is)=data(2,is)         ! y
         tiem2(3,is)=data(3,is)         ! H
         tiem2(4,is)=data(5,is)         ! D
         tiem2(5,is)=data(4,is)         ! Elev
         tiem2(6,is)=0.0                ! No BDW
         tiem2(7,is)=1.0                ! v. momentum factor
         tiem2(8,is)=0.0                ! user code
      enddo

      return
      end

c----------------------------------------------------------------------
      subroutine mk_case(iolst,action,cvar,ierr)
c----------------------------------------------------------------------
c
c --- CALPUFF    Version: 6.42_x1.0     Level: 121203           MK_CASE
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Imposes uniform case on a string
c
c --- INPUT:
c               IOLST - integer   - Unit number of output list file
c              ACTION - char*     - Action to perform
c                                   (UPPER or LOWER)
c
c --- OUTPUT:
c                IERR - integer   - Error flag (0 = No Error)
c
c --- INPUT/OUTPUT:
c                CVAR - char*     - String
c
c --- MK_CASE called by:  ()
c --- MK_CASE calls:      none
c----------------------------------------------------------------------
      implicit none

c --- Declare arguments
      character(len=*) :: cvar,action
      integer :: iolst,ierr

c --- Declare locals
      integer :: iul,lo(2),hi(2),shift(2)
      integer :: nc,ia,ic

c --- Define ASCII table pointers for conversion to UPPER and LOWER
c ---         (UPPER,LOWER)
      data lo   / 97,65/
      data hi   /122,90/
      data shift/-32,32/

c --- Initializations
      ierr=0

c --- Set conversion
      if(TRIM(action).EQ.'UPPER') then
         iul=1
      elseif(TRIM(action).EQ.'LOWER') then
         iul=2
      else
         write(iolst,*)'ERROR in MK_CASE -- Invalid ACTION'
         write(iolst,*)'   ACTION expected: UPPER or LOWER'
         write(iolst,*)'   ACTION found   : ',action
         ierr=1
         return
      endif

c --- String length
      nc=LEN(cvar)

c --- Process string
      do ic=1,nc
         ia=IACHAR(cvar(ic:ic))
         if(ia.GE.lo(iul) .AND. ia.LE.hi(iul)) then
            ia=ia+shift(iul)
            cvar(ic:ic)=ACHAR(ia)
         endif
      enddo

      return
      end
